import time
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import ConvLSTM2D, LSTM, Dense, Flatten, Dropout, \
                                    RepeatVector, TimeDistributed
# helper functions
from ..utils import print_dynamic_rmse, print_mae, print_mape


# def build_conv_model(X, y, n_input, n_output=1, n_features=1, n_units=64,
#                     d_rate=0.15, n_epoch=10, n_batch=1, verbose=0):
def build_convlstm_model(X_train, y_train, X_test, y_test, l_subseq=1, n_row=1, n_col=1,
                         n_features=1, n_units=64, d_rate=0.15, n_epoch=10, n_batch=1, verbose=0):

    # ConvLSTM Hyperparameters
    # [samples, subseq, rows, cols, channel]

    X_train = X_train.reshape((X_train.shape[0], int(X_train.shape[1]/l_subseq),
                               n_row, n_col, n_features))
    y_train = y_train.reshape((y_train.shape[0], n_col, n_features))
    X_test = X_test.reshape((X_test.shape[0], int(X_test.shape[1]/l_subseq),
                             n_row, n_col, n_features))
    y_test = y_test.reshape((y_test.shape[0], n_col, n_features))
    n_steps = X_train.shape[1]

    conv_model = StackedConvLSTM(n_steps, n_row, n_col, n_units, n_features, d_rate)
    conv_model.compile(optimizer="adam", loss="mse")
    start_time = time.time()
    conv_model.fit(X_train, y_train, epochs=n_epoch, batch_size=n_batch)
    end_time = time.time()

    conv_pred = conv_model.predict(X_test)
    rmse, norm_rmse = print_dynamic_rmse(y_test, conv_pred, y_train)
    mae = print_mae(y_test, conv_pred)
    mape = print_mape(y_test, conv_pred)
    run_time = end_time - start_time
    print("-----------------------------------------------------------------")
    print("ConvLSTM SUMMARY:")
    print("-----------------------------------------------------------------")
    print("X: [#samples, #subseq, #rows, #cols (outputs), #channel (features)]")
    print("y: [#samples, #cols (outputs), #channel (features)]")
    print('    Conv X_train.shape = ', X_train.shape)
    print('    Conv y_train.shape = ', y_train.shape)
    print('    Conv X_test.shape = ',  X_test.shape)
    print('    Conv y_test.shape = ', y_test.shape)
    print("-----------------------------------------------------------------")
    print(f"MAE Score: {round(mae, 4)}")
    print(f"MAPE Score: {round(mape, 4)}")
    print(f"RMSE Score: {round(rmse, 4)}")
    print(f"Normalized RMSE Score: {round(norm_rmse, 4)*100}%")
    print(f"Total Training Time: {round(run_time/60, 2)} min")

    return conv_model, conv_pred, rmse, norm_rmse


def StackedConvLSTM(n_steps, n_row, n_col, n_units, n_features, d_rate=0.5):
    model = Sequential()
    # define encoder
    model.add(ConvLSTM2D(64, (1,3), activation='relu', \
                         input_shape=(n_steps, n_row, n_col, n_features)))
    model.add(Flatten())
    # repeat encoding
    model.add(RepeatVector(n_col))
    # define decoder
    model.add(LSTM(n_units, activation='relu', return_sequences=True))
    # define output model
    model.add(TimeDistributed(Dense(n_units, activation='relu')))
    model.add(TimeDistributed(Dropout(d_rate)))
    model.add(TimeDistributed(Dense(n_features)))  # 1 unit if univariate time-series
    print("Stacked ConvLSTM model summary:")
    model.summary()
    return model


def CustomConvLSTM(n_input, n_output, n_units, n_features):
    pass
